#!/usr/bin/env python
import GuiTextArea
import F
from RouterSimulator import RouterSimulator
from RouterPacket import RouterPacket
from typing import List
from copy import deepcopy


class RouterNode():
    myID = None
    myGUI = None
    sim: RouterSimulator = None
    original_costs: List[int] = None
    sources: List[int] = None

    cost_table: List[List[int]] = None

    # Access simulator variables with:
    # self.sim.POISONREVERSE, self.sim.NUM_NODES, etc.

    # --------------------------------------------------
    def __init__(self, ID, sim: RouterSimulator, costs: List[int]):
        self.new_data_to_print = True
        self.myID = ID
        self.sim = sim
        self.myGUI = GuiTextArea.GuiTextArea(
            "  Output window for Router #" + str(ID) + "  ")

        self.original_costs = deepcopy(costs)

        self.cost_table = []
        self.sources = []
        for i in range(self.sim.NUM_NODES):
            self.cost_table.append([sim.INFINITY] * sim.NUM_NODES)
            self.sources.append(i)
        
        self.cost_table[self.myID] = deepcopy(costs)

        self.sendUpdateToAll()

    # --------------------------------------------------

    def update_cost_table(self, changed=False):
        my_costs = self.cost_table[self.myID]

        for node_id in range(self.sim.NUM_NODES):

            for neighbor_id in range(self.sim.NUM_NODES):

                if neighbor_id == self.myID:
                    continue

                new_cost = self.cost_table[node_id][neighbor_id] + self.original_costs[node_id]
                neighbor_cost = my_costs[neighbor_id]

                if new_cost < neighbor_cost or (self.sources[neighbor_id] == node_id and neighbor_cost != new_cost):
                    my_costs[neighbor_id] = new_cost
                    self.sources[neighbor_id] = node_id
                    changed = True

        # If cost has changed
        if changed:
            self.new_data_to_print = True
            self.sendUpdateToAll()

    # --------------------------------------------------

    def recvUpdate(self, pkt: RouterPacket):
        # Update table for source node
        self.cost_table[pkt.sourceid] = pkt.mincost

        self.update_cost_table()

    # --------------------------------------------------

    def sendUpdateToAll(self):
        for id in range(self.sim.NUM_NODES):
            if id == self.myID:
                continue
 
            tmp = deepcopy(self.cost_table[self.myID])
            if self.sim.POISONREVERSE:
                """ Poison reverse """
                for n in range(self.sim.NUM_NODES):
                    if id == self.sources[n]:
                        tmp[n] = self.sim.INFINITY

            pkt = RouterPacket(self.myID, id, tmp)
            self.sendUpdate(pkt)

    # --------------------------------------------------

    def sendUpdate(self, pkt: RouterPacket):
        self.sim.toLayer2(pkt)

    # --------------------------------------------------

    def printDistanceTable(self):
        if not self.new_data_to_print:
            return

        self.new_data_to_print = False
        self.myGUI.println("Current table for " + str(self.myID) +
                           "  at time " + str(self.sim.getClocktime()))
        self.myGUI.println("Costs: " + str(self.cost_table[self.myID]))
        self.myGUI.println("Sources: " + str(self.sources))

    # --------------------------------------------------

    def updateLinkCost(self, dest, newcost):
        self.original_costs[dest] = newcost
        my_costs = self.cost_table[self.myID]

        if dest == self.sources[dest] or newcost < my_costs[dest]:
            my_costs[dest] = newcost
            self.sources[dest] = dest

            self.new_data_to_print = True

            self.update_cost_table(changed=True)
